# -*- coding: utf-8 -*-
# Copyright (C) 2014-2017 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
import re
from types import FunctionType
import warnings
import sys

from ..i18n import _
from ..setting import log, undefined, debug
from ..autolib import carry_out_calculation
from ..error import (ConfigError, ValueWarning, PropertiesOptionError,
                     display_list)

if sys.version_info[0] >= 3:  # pragma: no cover
    from inspect import signature
else:
    from inspect import getargspec

static_tuple = tuple()

if sys.version_info[0] >= 3:  # pragma: no cover
    xrange = range


submulti = 2
name_regexp = re.compile(r'^[a-z][a-zA-Z\d_]*$')
FORBIDDEN_NAMES = frozenset(['iter_all', 'iter_group', 'find', 'find_first',
                             'make_dict', 'unwrap_from_path', 'read_only',
                             'read_write', 'getowner', 'set_contexts'])
ALLOWED_CONST_LIST = ['_cons_not_equal']


def valid_name(name):
    "an option's name is a str and does not start with 'impl' or 'cfgimpl'"
    if not isinstance(name, str):
        return False
    if re.match(name_regexp, name) is not None and \
            name not in FORBIDDEN_NAMES and \
            not name.startswith('impl_') and \
            not name.startswith('cfgimpl_'):
        return True
    else:
        return False


def validate_callback(callback, callback_params, type_, callbackoption):
    if not isinstance(callback, FunctionType):
        raise ValueError(_('{0} must be a function').format(type_))
    if callback_params is not None:
        if not isinstance(callback_params, dict):
            raise ValueError(_('{0}_params must be a dict').format(type_))
        for key, callbacks in callback_params.items():
            if key != '' and len(callbacks) != 1:
                raise ValueError(_("{0}_params with key {1} mustn't have "
                                   "length different to 1").format(type_,
                                                                   key))
            if not isinstance(callbacks, tuple):
                raise ValueError(_('{0}_params must be tuple for key "{1}"'
                                  ).format(type_, key))
            for callbk in callbacks:
                if isinstance(callbk, tuple):
                    if len(callbk) == 1:
                        if callbk not in  ((None,), ('index',)):
                            raise ValueError(_('{0}_params with length of '
                                               'tuple as 1 must only have '
                                               'None as first value'))
                    elif len(callbk) != 2:
                        raise ValueError(_('{0}_params must only have 1 or 2 '
                                           'as length'))
                    else:
                        option, force_permissive = callbk
                        if not isinstance(option, Option) and not \
                                isinstance(option, SymLinkOption):
                            raise ValueError(_('{}_params must have an option'
                                               ' not a {} for first argument'
                                              ).format(type_, type(option)))
                        if force_permissive not in [True, False]:
                            raise ValueError(_('{}_params must have a boolean'
                                               ' not a {} for second argument'
                                              ).format(type_, type(
                                                  force_permissive)))
                        if isinstance(option, SymLinkOption):
                            cur_opt = option._impl_getopt()
                        else:
                            cur_opt = option
                        if cur_opt != callbackoption:
                            if not getattr(cur_opt, '_dependencies', None):
                                options = set()
                            else:
                                options = set(cur_opt._dependencies)
                            options.add(callbackoption)
                            cur_opt._dependencies = tuple(options)
#____________________________________________________________
#


class Base(object):
    __slots__ = ('_name',
                 '_informations',
                 '_extra',
                 '_warnings_only',
                 '_allow_empty_list',
                 #multi
                 '_multi',
                 '_unique',
                 #value
                 '_default',
                 '_default_multi',
                 #calcul
                 '_subdyn',
                 '_requires',
                 '_properties',
                 '_calc_properties',
                 '_val_call',
                 #
                 '_consistencies',
                 '_master_slaves',
                 '_choice_values',
                 '_choice_values_params',
                 #other
                 '_has_dependency',
                 '_dependencies',
                 '__weakref__'
                )

    def __init__(self, name, doc, default=None, default_multi=None,
                 requires=None, multi=False, unique=undefined, callback=None,
                 callback_params=None, validator=None, validator_params=None,
                 properties=None, warnings_only=False, extra=None,
                 allow_empty_list=undefined):
        if not valid_name(name):
            raise ValueError(_("invalid name: {0} for option").format(name))
        if not multi and default_multi is not None:
            raise ValueError(_("default_multi is set whereas multi is False"
                             " in option: {0}").format(name))
        if multi is True:
            is_multi = True
            _multi = 0
        elif multi is False:
            is_multi = False
            _multi = 1
        elif multi is submulti:
            is_multi = True
            _multi = submulti
        else:
            raise ValueError(_('invalid multi value'))
        if unique != undefined and not isinstance(unique, bool):
            raise ValueError(_('unique must be a boolean'))
        if not is_multi and unique is True:
            raise ValueError(_('unique must be set only with multi value'))
        if requires is not None:
            calc_properties, requires = validate_requires_arg(self, is_multi,
                                                              requires, name)
        else:
            calc_properties = frozenset()
            requires = undefined
        if properties is None:
            properties = tuple()
        if not isinstance(properties, tuple):
            raise TypeError(_('invalid properties type {0} for {1},'
                              ' must be a tuple').format(
                                  type(properties),
                                  name))
        if validator is not None:
            if multi:  # and validator_params is None:
                validator_params = self._build_validator_params(validator, validator_params)

            validate_callback(validator, validator_params, 'validator', self)
            if validator_params is None:
                val_call = (validator,)
            else:
                val_call = (validator, validator_params)
            self._val_call = (val_call, None)
            self._set_has_dependency()
        if calc_properties != frozenset([]) and properties is not tuple():
            set_forbidden_properties = calc_properties & set(properties)
            if set_forbidden_properties != frozenset():
                raise ValueError('conflict: properties already set in '
                                 'requirement {0}'.format(
                                     list(set_forbidden_properties)))
        _setattr = object.__setattr__
        _setattr(self, '_name', name)
        if sys.version_info[0] < 3 and isinstance(doc, str):
            doc = doc.decode('utf8')
        if extra is not None:
            _setattr(self, '_extra', extra)
        _setattr(self, '_informations', {'doc': doc})
        if _multi != 1:
            _setattr(self, '_multi', _multi)
        if warnings_only is True:
            _setattr(self, '_warnings_only', warnings_only)
        if calc_properties is not undefined:
            _setattr(self, '_calc_properties', calc_properties)
        if requires is not undefined:
            _setattr(self, '_requires', requires)
        if properties is not undefined:
            _setattr(self, '_properties', properties)
        if multi is not False and default is None:
            default = []
        if allow_empty_list is not undefined:
            _setattr(self, '_allow_empty_list', allow_empty_list)
        if unique is not undefined:
            _setattr(self, '_unique', unique)
        err = self.impl_validate(default, is_multi=is_multi)
        if err:
            raise err
        if (is_multi and default != []) or \
                (not is_multi and default is not None):
            if is_multi:
                default = tuple(default)
            _setattr(self, '_default', default)

        if is_multi and default_multi is not None:
            err = self._validate(default_multi)
            if err:
                raise ValueError(_("invalid default_multi value {0} "
                                   "for option {1}: {2}").format(
                                       str(default_multi),
                                       self.impl_getname(), str(err)))
            _setattr(self, '_default_multi', default_multi)

        ##callback is False in optiondescription
        if callback is not False:
            self.impl_set_callback(callback, callback_params, _init=True)

    def _build_validator_params(self, validator, validator_params):
        if sys.version_info[0] < 3:
            func_args = getargspec(validator)
            defaults = func_args.defaults
            if defaults is None:
                defaults = []
            args = func_args.args[0:len(func_args.args)-len(defaults)]
        else:  # pragma: no cover
            func_params = signature(validator).parameters
            args = [f.name for f in func_params.values() if f.default is f.empty]
        if validator_params is not None:
            kwargs = list(validator_params.keys())
            if '' in kwargs:
                kwargs.remove('')
            for kwarg in kwargs:
                if kwarg in args:
                    args = args[0:args.index(kwarg)]
            len_args = len(validator_params.get('', []))
            if len_args != 0 and len(args) >= len_args:
                args = args[0:len(args)-len_args]
        if len(args) >= 2:
            if validator_params is not None and '' in validator_params:
                params = list(validator_params[''])
                params.append((self, False))
                validator_params[''] = tuple(params)
            else:
                if validator_params is None:
                    validator_params = {}
                validator_params[''] = ((self, False),)
        if len(args) == 3 and args[2] not in validator_params:
            params = list(validator_params[''])
            params.append(('index',))
            validator_params[''] = tuple(params)
        return validator_params

    def _set_has_dependency(self):
        if not isinstance(self, SymLinkOption):
            self._has_dependency = True

    def impl_has_dependency(self):
        return getattr(self, '_has_dependency', False)

    def impl_set_callback(self, callback, callback_params=None, _init=False):
        if callback is None and callback_params is not None:
                raise ValueError(_("params defined for a callback function but "
                                   "no callback defined"
                                   " yet for option {0}").format(
                                       self.impl_getname()))
        if not _init and self.impl_get_callback()[0] is not None:
            raise ConfigError(_("a callback is already set for {0}, "
                                "cannot set another one's").format(self.impl_getname()))
        self._validate_callback(callback, callback_params)
        if callback is not None:
            validate_callback(callback, callback_params, 'callback', self)
            val = getattr(self, '_val_call', (None,))[0]
            if callback_params is None or callback_params == {}:
                val_call = (callback,)
            else:
                val_call = tuple([callback, callback_params])
            self._val_call = (val, val_call)

    def impl_is_optiondescription(self):
        return self.__class__.__name__ in ['OptionDescription',
                                           'DynOptionDescription',
                                           'SynDynOptionDescription']

    def impl_is_dynoptiondescription(self):
        return self.__class__.__name__ in ['DynOptionDescription',
                                           'SynDynOptionDescription']

    def impl_getname(self):
        return self._name

    def impl_is_multi(self):
        return getattr(self, '_multi', 1) != 1

    def impl_is_readonly(self):
        return not isinstance(getattr(self, '_informations', dict()), dict)

    def impl_getproperties(self):
        return self._properties

    def _set_readonly(self, has_extra):
        if not self.impl_is_readonly():
            _setattr = object.__setattr__
            dico = self._informations
            keys = tuple(dico.keys())
            if len(keys) == 1:
                dico = dico['doc']
            else:
                dico = tuple([keys, tuple(dico.values())])
            _setattr(self, '_informations', dico)
            if has_extra:
                extra = getattr(self, '_extra', None)
                if extra is not None:
                    _setattr(self, '_extra', tuple([tuple(extra.keys()), tuple(extra.values())]))

    def _impl_setsubdyn(self, subdyn):
        self._subdyn = subdyn

    def impl_getrequires(self):
        return getattr(self, '_requires', static_tuple)

    def impl_get_callback(self):
        call = getattr(self, '_val_call', (None, None))[1]
        if call is None:
            ret_call = (None, {})
        elif len(call) == 1:
            ret_call = (call[0], {})
        else:
            ret_call = call
        return ret_call

    # ____________________________________________________________
    # information
    def impl_get_information(self, key, default=undefined):
        """retrieves one information's item

        :param key: the item string (ex: "help")
        """
        def _is_string(infos):
            if sys.version_info[0] >= 3:  # pragma: no cover
                return isinstance(infos, str)
            else:
                return isinstance(infos, str) or isinstance(infos, unicode)

        dico = self._informations
        if isinstance(dico, tuple):
            if key in dico[0]:
                return dico[1][dico[0].index(key)]
        elif _is_string(dico):
            if key == 'doc':
                return dico
        elif isinstance(dico, dict):
            if key in dico:
                return dico[key]
        if default is not undefined:
            return default
        raise ValueError(_("information's item not found: {0}").format(
            key))

    def impl_set_information(self, key, value):
        """updates the information's attribute
        (which is a dictionary)

        :param key: information's key (ex: "help", "doc"
        :param value: information's value (ex: "the help string")
        """
        if self.impl_is_readonly():
            raise AttributeError(_("'{0}' ({1}) object attribute '{2}' is"
                                   " read-only").format(
                                       self.__class__.__name__,
                                       self,
                                       #self.impl_getname(),
                                       key))
        self._informations[key] = value


class BaseOption(Base):
    """This abstract base class stands for attribute access
    in options that have to be set only once, it is of course done in the
    __setattr__ method
    """
    __slots__ = tuple()

    def __getstate__(self):
        raise NotImplementedError()

    def __setattr__(self, name, value):
        """set once and only once some attributes in the option,
        like `_name`. `_name` cannot be changed one the option and
        pushed in the :class:`tiramisu.option.OptionDescription`.

        if the attribute `_readonly` is set to `True`, the option is
        "frozen" (which has noting to do with the high level "freeze"
        propertie or "read_only" property)
        """
        if name != '_option' and \
                not isinstance(value, tuple):
            is_readonly = False
            # never change _name dans _opt
            if name == '_name':
                if self.impl_getname() is not None:
                    #so _name is already set
                    is_readonly = True
            elif name != '_readonly':
                is_readonly = self.impl_is_readonly()
            if is_readonly:
                raise AttributeError(_("'{0}' ({1}) object attribute '{2}' is"
                                       " read-only").format(
                                           self.__class__.__name__,
                                           self,
                                           #self.impl_getname(),
                                           name))
        super(BaseOption, self).__setattr__(name, value)

    def impl_getpath(self, context):
        return context.cfgimpl_get_description().impl_get_path_by_opt(self)

    def impl_has_callback(self):
        "to know if a callback has been defined or not"
        return self.impl_get_callback()[0] is not None

    def _is_subdyn(self):
        return getattr(self, '_subdyn', None) is not None

    def _impl_valid_unicode(self, value):
        if sys.version_info[0] >= 3:  # pragma: no cover
            if not isinstance(value, str):
                return ValueError(_('invalid string'))
        else:
            if not isinstance(value, unicode) and not isinstance(value, str):
                return ValueError(_('invalid unicode or string'))

    def impl_get_display_name(self, dyn_name=None):
        name = self.impl_getdoc()
        if name is None or name == '':
            if dyn_name is not None:
                name = dyn_name
            else:
                name = self.impl_getname()
        if sys.version_info[0] < 3 and isinstance(name, unicode):
            name = name.encode('utf8')
        return name

    def reset_cache(self, opt, obj, type_):
        context = obj._getcontext()
        path = self.impl_getpath(context)
        obj._p_.delcache(path)
        context.cfgimpl_reset_cache(only=(type_,),
                                    opt=self,
                                    path=path)


class OnlyOption(BaseOption):
    __slots__ = tuple()


class Option(OnlyOption):
    """
    Abstract base class for configuration option's.

    Reminder: an Option object is **not** a container for the value.
    """
    __slots__ = tuple()
    _empty = ''

    def _launch_consistency(self, current_opt, func, option, value, context,
                            index, submulti_index, opts, warnings_only,
                            transitive):
        """Launch consistency now

        :param func: function name, this name should start with _cons_
        :type func: `str`
        :param option: option that value is changing
        :type option: `tiramisu.option.Option`
        :param value: new value of this option
        :param context: Config's context, if None, check default value instead
        :type context: `tiramisu.config.Config`
        :param index: only for multi option, consistency should be launch for
                      specified index
        :type index: `int`
        :param opts: all options concerne by this consistency
        :type opts: `list` of `tiramisu.option.Option`
        :param warnings_only: specific raise error for warning
        :type warnings_only: `boolean`
        :param transitive: propertyerror is transitive
        :type transitive: `boolean`
        """
        if context is not undefined:
            descr = context.cfgimpl_get_description()

        all_cons_vals = []
        all_cons_opts = []
        val_consistencies = True
        for opt in opts:
            if (isinstance(opt, DynSymLinkOption) and option._dyn == opt._dyn) or \
                    option == opt:
                # option is current option
                # we have already value, so use it
                all_cons_vals.append(value)
                all_cons_opts.append(opt)
            else:
                #if context, calculate value, otherwise get default value
                path = None
                is_multi = opt.impl_is_multi() and not opt.impl_is_master_slaves()
                if context is not undefined:
                    if isinstance(opt, DynSymLinkOption):
                        path = opt.impl_getpath(context)
                    else:
                        path = descr.impl_get_path_by_opt(opt)
                    if is_multi:
                        _index = None
                    else:
                        _index = index
                    opt_value = context.getattr(path, validate=False,
                                                index=_index,
                                                force_permissive=True,
                                                returns_raise=True)
                    if isinstance(opt_value, Exception):
                        if isinstance(opt_value, PropertiesOptionError):
                            if debug:  # pragma: no cover
                                log.debug('propertyerror in _launch_consistency: {0}'.format(opt_value))
                            if transitive:
                                opt_value.set_orig_opt(option)
                                return opt_value
                            else:
                                opt_value = None
                        else:  # pragma: no cover
                            return opt_value
                elif index is None:
                    opt_value = opt.impl_getdefault()
                else:
                    opt_value = opt.impl_getdefault()[index]

                if self.impl_is_multi() and index is None:
                    # only check propertyerror for master/slaves is transitive
                    val_consistencies = False
                if is_multi and isinstance(opt_value, list):
                    all_cons_vals.extend(opt_value)
                    for len_ in xrange(len(opt_value)):
                        all_cons_opts.append(opt)
                else:
                    all_cons_vals.append(opt_value)
                    all_cons_opts.append(opt)

        if val_consistencies:
            err = getattr(self, func)(current_opt, all_cons_opts, all_cons_vals, warnings_only)
            if err:
                if warnings_only:
                    msg = _('attention, "{0}" could be an invalid {1} for "{2}", {3}').format(
                        value, self._display_name, current_opt.impl_get_display_name(), err)
                    warnings.warn_explicit(ValueWarning(msg, self),
                                           ValueWarning,
                                           self.__class__.__name__, 0)
                else:
                    return err

    def impl_is_unique(self):
        return getattr(self, '_unique', False)

    def impl_get_validator(self):
        val = getattr(self, '_val_call', (None,))[0]
        if val is None:
            ret_val = (None, {})
        elif len(val) == 1:
            ret_val = (val[0], {})
        else:
            ret_val = val
        return ret_val

    def impl_validate(self, value, context=undefined, validate=True,
                      force_index=None, force_submulti_index=None,
                      current_opt=undefined, is_multi=None,
                      display_error=True, display_warnings=True, multi=None,
                      setting_properties=undefined):
        """
        :param value: the option's value
        :param context: Config's context
        :type context: :class:`tiramisu.config.Config`
        :param validate: if true enables ``self._validator`` validation
        :type validate: boolean
        :param force_index: if multi, value has to be a list
                            not if force_index is not None
        :type force_index: integer
        :param force_submulti_index: if submulti, value has to be a list
                                     not if force_submulti_index is not None
        :type force_submulti_index: integer
        """
        if not validate:
            return
        if current_opt is undefined:
            current_opt = self

        if display_warnings and setting_properties is undefined and context is not undefined:
            setting_properties = context.cfgimpl_get_settings()._getproperties(read_write=False)
        display_warnings = display_warnings and (setting_properties is undefined or 'warnings' in setting_properties)

        def _is_not_unique(value):
            if display_error and self.impl_is_unique() and len(set(value)) != len(value):
                for idx, val in enumerate(value):
                    if val in value[idx+1:]:
                        return ValueError(_('invalid value "{}", this value is already in "{}"').format(
                                            val, self.impl_get_display_name()))

        def calculation_validator(val, _index):
            validator, validator_params = self.impl_get_validator()
            if validator is not None:
                if validator_params != {}:
                    validator_params_ = {}
                    for val_param, values in validator_params.items():
                        validator_params_[val_param] = values
                    #inject value in calculation
                    if '' in validator_params_:
                        lst = list(validator_params_[''])
                        lst.insert(0, val)
                        validator_params_[''] = tuple(lst)
                    else:
                        validator_params_[''] = (val,)
                else:
                    validator_params_ = {'': (val,)}
                # Raise ValueError if not valid
                value = carry_out_calculation(current_opt, context=context,
                                              callback=validator,
                                              callback_params=validator_params_,
                                              index=_index,
                                              is_validator=True)
                if isinstance(value, Exception):
                    return value

        def do_validation(_value, _index, submulti_index):
            if _value is None:
                error = warning = None
            else:
                if display_error:
                    # option validation
                    err = self._validate(_value, context, current_opt)
                    if err:
                        if debug:  # pragma: no cover
                            log.debug('do_validation: value: {0}, index: {1}, '
                                      'submulti_index: {2}'.format(_value, _index,
                                                                   submulti_index),
                                      exc_info=True)
                        err_msg = '{0}'.format(err)
                        if err_msg:
                            msg = _('"{0}" is an invalid {1} for "{2}", {3}'
                                    '').format(_value, self._display_name,
                                               self.impl_get_display_name(), err_msg)
                        else:
                            msg = _('"{0}" is an invalid {1} for "{2}"'
                                    '').format(_value, self._display_name,
                                               self.impl_get_display_name())
                        return ValueError(msg)
                error = None
                is_warnings_only = getattr(self, '_warnings_only', False)
                if ((display_error and not is_warnings_only) or
                        (display_warnings and is_warnings_only)):
                    error = calculation_validator(_value, _index)
                    if not error:
                        error = self._second_level_validation(_value, is_warnings_only)
                    if error:
                        if debug:  # pragma: no cover
                            log.debug(_('do_validation for {0}: error in value').format(
                                self.impl_getname()), exc_info=True)
                        if is_warnings_only:
                            msg = _('attention, "{0}" could be an invalid {1} for "{2}", {3}').format(
                                _value, self._display_name, self.impl_get_display_name(), error)
                            warnings.warn_explicit(ValueWarning(msg, self),
                                                   ValueWarning,
                                                   self.__class__.__name__, 0)
                            error = None
            if error is None:
                # if context launch consistency validation
                #if context is not undefined:
                ret = self._valid_consistency(current_opt, _value, context,
                                              _index, submulti_index, display_warnings,
                                              display_error)
                if isinstance(ret, ValueError):
                    error = ret
                elif ret:
                    return ret
            if error:
                err_msg = '{0}'.format(error)
                if err_msg:
                    msg = _('"{0}" is an invalid {1} for "{2}", {3}'
                            '').format(_value, self._display_name,
                                       self.impl_get_display_name(), err_msg)
                else:
                    msg = _('"{0}" is an invalid {1} for "{2}"'
                            '').format(_value, self._display_name,
                                       self.impl_get_display_name())
                return ValueError(msg)

        if is_multi is None:
            is_multi = self.impl_is_multi()

        if not is_multi:
            return do_validation(value, None, None)
        elif force_index is not None:
            if self.impl_is_submulti() and force_submulti_index is None:
                err = _is_not_unique(value)
                if err:
                    return err
                if not isinstance(value, list):
                    return ValueError(_('invalid value "{0}" for "{1}" which'
                                        ' must be a list').format(
                                           value, self.impl_get_display_name()))
                for idx, val in enumerate(value):
                    if isinstance(val, list):  # pragma: no cover
                        return ValueError(_('invalid value "{}" for "{}" '
                                            'which must not be a list').format(val,
                                                                              self.impl_get_display_name()))
                    err = do_validation(val, force_index, idx)
                    if err:
                        return err
            else:
                if multi is not None and self.impl_is_unique() and value in multi:
                    if not self.impl_is_submulti() and len(multi) - 1 >= force_index:
                        lst = list(multi)
                        lst.pop(force_index)
                    else:
                        lst = multi
                    if value in lst:
                        return ValueError(_('invalid value "{}", this value is already'
                                            ' in "{}"').format(value,
                                                               self.impl_get_display_name()))
                return do_validation(value, force_index, force_submulti_index)
        elif not isinstance(value, list):
            return ValueError(_('invalid value "{0}" for "{1}" which '
                                'must be a list').format(value,
                                                         self.impl_getname()))
        elif self.impl_is_submulti() and force_submulti_index is None:
            for idx, val in enumerate(value):
                err = _is_not_unique(val)
                if err:
                    return err
                if not isinstance(val, list):
                    return ValueError(_('invalid value "{0}" for "{1}" '
                                        'which must be a list of list'
                                        '').format(val,
                                                   self.impl_getname()))
                for slave_idx, slave_val in enumerate(val):
                    err = do_validation(slave_val, idx, slave_idx)
                    if err:
                        return err
        else:
            err = _is_not_unique(value)
            if err:
                return err
            for idx, val in enumerate(value):
                err = do_validation(val, idx, force_submulti_index)
                if err:
                    return err
            return self._valid_consistency(current_opt, None, context,
                                           None, None, display_warnings, display_error)

    def impl_is_dynsymlinkoption(self):
        return False

    def impl_is_master_slaves(self, type_='both'):
        """FIXME
        """
        master_slaves = self.impl_get_master_slaves()
        if master_slaves is not None:
            if type_ in ('both', 'master') and \
                    master_slaves.is_master(self):
                return True
            if type_ in ('both', 'slave') and \
                    not master_slaves.is_master(self):
                return True
        return False

    def impl_get_master_slaves(self):
        return getattr(self, '_master_slaves', None)

    def impl_getdoc(self):
        "accesses the Option's doc"
        return self.impl_get_information('doc')

    def _valid_consistencies(self, other_opts, init=True, func=None):
        if self._is_subdyn():
            dynod = self._subdyn
        else:
            dynod = None
        if self.impl_is_submulti():
            raise ConfigError(_('cannot add consistency with submulti option'))
        is_multi = self.impl_is_multi()
        for opt in other_opts:
            if opt.impl_is_submulti():
                raise ConfigError(_('cannot add consistency with submulti option'))
            if not isinstance(opt, Option):
                raise ConfigError(_('consistency must be set with an option'))
            if opt._is_subdyn():
                if dynod is None:
                    raise ConfigError(_('almost one option in consistency is '
                                        'in a dynoptiondescription but not all'))
                if dynod != opt._subdyn:
                    raise ConfigError(_('option in consistency must be in same'
                                        ' dynoptiondescription'))
                dynod = opt._subdyn
            elif dynod is not None:
                raise ConfigError(_('almost one option in consistency is in a '
                                    'dynoptiondescription but not all'))
            if self is opt:
                raise ConfigError(_('cannot add consistency with itself'))
            if is_multi != opt.impl_is_multi():
                raise ConfigError(_('every options in consistency must be '
                                    'multi or none'))
            if init:
                # FIXME
                if func != 'not_equal':
                    opt._set_has_dependency()

    def impl_add_consistency(self, func, *other_opts, **params):
        """Add consistency means that value will be validate with other_opts
        option's values.

        :param func: function's name
        :type func: `str`
        :param other_opts: options used to validate value
        :type other_opts: `list` of `tiramisu.option.Option`
        :param params: extra params (warnings_only and transitive are allowed)
        """
        if self.impl_is_readonly(): 
            raise AttributeError(_("'{0}' ({1}) cannot add consistency, option is"
                                   " read-only").format(
                                       self.__class__.__name__,
                                       self.impl_getname()))
        self._valid_consistencies(other_opts, func=func)
        func = '_cons_{0}'.format(func)
        if func not in dir(self):
            raise ConfigError(_('consistency {0} not available for this option').format(func))
        all_cons_opts = tuple([self] + list(other_opts))
        unknown_params = set(params.keys()) - set(['warnings_only', 'transitive'])
        if unknown_params != set():
            raise ValueError(_('unknow parameter {0} in consistency').format(unknown_params))
        self._add_consistency(func, all_cons_opts, params)
        #validate default value when add consistency
        err = self.impl_validate(self.impl_getdefault())
        if err:
            self._del_consistency()
            raise err
        if func in ALLOWED_CONST_LIST:
            for opt in all_cons_opts:
                if getattr(opt, '_unique', undefined) == undefined:
                    opt._unique = True
        if func != '_cons_not_equal':
            #consistency could generate warnings or errors
            self._set_has_dependency()

    def _valid_consistency(self, option, value, context, index, submulti_idx,
                           display_warnings, display_error):
        if context is not undefined:
            descr = context.cfgimpl_get_description()
            if descr._cache_consistencies is None:
                return
            #consistencies is something like [('_cons_not_equal', (opt1, opt2))]
            if isinstance(option, DynSymLinkOption):
                consistencies = descr._cache_consistencies.get(option._impl_getopt())
            else:
                consistencies = descr._cache_consistencies.get(option)
        else:
            consistencies = option._get_consistencies()
        if consistencies is not None:
            for func, all_cons_opts, params in consistencies:
                warnings_only = params.get('warnings_only', False)
                if (warnings_only and display_warnings) or (not warnings_only and display_error):
                    transitive = params.get('transitive', True)
                    #all_cons_opts[0] is the option where func is set
                    if isinstance(option, DynSymLinkOption):
                        subpath = '.'.join(option._dyn.split('.')[:-1])
                        namelen = len(option._impl_getopt().impl_getname())
                        suffix = option.impl_getname()[namelen:]
                        opts = []
                        for opt in all_cons_opts:
                            name = opt.impl_getname() + suffix
                            path = subpath + '.' + name
                            opts.append(opt._impl_to_dyn(name, path))
                    else:
                        opts = all_cons_opts
                    err = opts[0]._launch_consistency(self, func, option, value,
                                                      context, index, submulti_idx,
                                                      opts, warnings_only,
                                                      transitive)
                    if err:
                        return err

    def _cons_not_equal(self, current_opt, opts, vals, warnings_only):
        equal = set()
        is_current = False
        for idx_inf, val_inf in enumerate(vals):
            for idx_sup, val_sup in enumerate(vals[idx_inf + 1:]):
                if val_inf == val_sup is not None:
                    for opt_ in [opts[idx_inf], opts[idx_inf + idx_sup + 1]]:
                        if opt_ == current_opt:
                            is_current = True
                        else:
                            equal.add(opt_)
        if equal:
            if debug:  # pragma: no cover
                log.debug(_('_cons_not_equal: {} are not different').format(display_list(list(equal))))
            if is_current:
                if warnings_only:
                    msg = _('should be different from the value of {}')
                else:
                    msg = _('must be different from the value of {}')
            else:
                if warnings_only:
                    msg = _('value for {} should be different')
                else:
                    msg = _('value for {} must be different')
            equal_name = []
            for opt in equal:
                equal_name.append(opt.impl_get_display_name())
            return ValueError(msg.format(display_list(list(equal_name))))

    def _second_level_validation(self, value, warnings_only):
        pass

    def _impl_to_dyn(self, name, path):
        return DynSymLinkOption(name, self, dyn=path)

    def impl_getdefault_multi(self):
        "accessing the default value for a multi"
        return getattr(self, '_default_multi', None)

    def _validate_callback(self, callback, callback_params):
        """callback_params:
        * None
        * {'': ((option, permissive),), 'ip': ((None,), (option, permissive))
        """
        if callback is None:
            return
        default_multi = self.impl_getdefault_multi()
        is_multi = self.impl_is_multi()
        default = self.impl_getdefault()
        if (not is_multi and (default is not None or default_multi is not None)) or \
                (is_multi and (default != [] or default_multi is not None)):
            raise ValueError(_("default value not allowed if option: {0} "
                             "is calculated").format(self.impl_getname()))

    def impl_getdefault(self):
        "accessing the default value"
        is_multi = self.impl_is_multi()
        default = getattr(self, '_default', undefined)
        if default is undefined:
            if is_multi:
                default = []
            else:
                default = None
        else:
            if is_multi:
                default = list(default)
        return default

    def _get_extra(self, key):
        extra = self._extra
        if isinstance(extra, tuple):
            return extra[1][extra[0].index(key)]
        else:
            return extra[key]

    def impl_is_submulti(self):
        return getattr(self, '_multi', 1) == 2

    def impl_allow_empty_list(self):
        return getattr(self, '_allow_empty_list', undefined)

    #____________________________________________________________
    # consistency
    def _add_consistency(self, func, all_cons_opts, params):
        cons = (func, all_cons_opts, params)
        consistencies = getattr(self, '_consistencies', None)
        if consistencies is None:
            self._consistencies = [cons]
        else:
            consistencies.append(cons)

    def _del_consistency(self):
        self._consistencies.pop(-1)

    def _get_consistencies(self):
        return getattr(self, '_consistencies', static_tuple)

    def _has_consistencies(self):
        return hasattr(self, '_consistencies')


def validate_requires_arg(new_option, multi, requires, name):
    """check malformed requirements
    and tranform dict to internal tuple

    :param requires: have a look at the
                     :meth:`tiramisu.setting.Settings.apply_requires` method to
                     know more about
                     the description of the requires dictionary
    """
    def set_dependency(option):
        if not getattr(option, '_dependencies', None):
            options = set()
        else:
            options = set(option._dependencies)
        options.add(new_option)
        option._dependencies = tuple(options)

    def get_option(require):
        option = require['option']
        if not isinstance(option, Option):
            raise ValueError(_('malformed requirements '
                               'must be an option in option {0}').format(name))
        if not multi and option.impl_is_multi():
            raise ValueError(_('malformed requirements '
                               'multi option must not set '
                               'as requires of non multi option {0}').format(name))
        set_dependency(option)
        return option

    def _set_expected(action, inverse, transitive, same_action, option, expected, operator):
        if inverse not in ret_requires[action]:
            ret_requires[action][inverse] = ([(option, [expected])], action, inverse, transitive, same_action, operator)
        else:
            for exp in ret_requires[action][inverse][0]:
                if exp[0] == option:
                    exp[1].append(expected)
                    break
            else:
                ret_requires[action][inverse][0].append((option, [expected]))

    def set_expected(require, ret_requires):
        expected = require['expected']
        inverse = get_inverse(require)
        transitive = get_transitive(require)
        same_action = get_sameaction(require)
        operator = get_operator(require)
        if isinstance(expected, list):
            for exp in expected:
                if set(exp.keys()) != {'option', 'value'}:
                    raise ValueError(_('malformed requirements expected must have '
                                       'option and value for option {0}').format(name))
                option = exp['option']
                set_dependency(option)
                if option is not None:
                    err = option._validate(exp['value'])
                    if err:
                        raise ValueError(_('malformed requirements expected value '
                                           'must be valid for option {0}'
                                           ': {1}').format(name, err))
                _set_expected(action, inverse, transitive, same_action, option, exp['value'], operator)
        else:
            option = get_option(require)
            if expected is not None:
                err = option._validate(expected)
                if err:
                    raise ValueError(_('malformed requirements expected value '
                                       'must be valid for option {0}'
                                       ': {1}').format(name, err))
            _set_expected(action, inverse, transitive, same_action, option, expected, operator)

    def get_action(require):
        action = require['action']
        if action == 'force_store_value':
            raise ValueError(_("malformed requirements for option: {0}"
                               " action cannot be force_store_value"
                               ).format(name))
        return action

    def get_inverse(require):
        inverse = require.get('inverse', False)
        if inverse not in [True, False]:
            raise ValueError(_('malformed requirements for option: {0}'
                               ' inverse must be boolean'))
        return inverse

    def get_transitive(require):
        transitive = require.get('transitive', True)
        if transitive not in [True, False]:
            raise ValueError(_('malformed requirements for option: {0}'
                               ' transitive must be boolean'))
        return transitive

    def get_sameaction(require):
        same_action = require.get('same_action', True)
        if same_action not in [True, False]:
            raise ValueError(_('malformed requirements for option: {0}'
                               ' same_action must be boolean'))
        return same_action

    def get_operator(require):
        operator = require.get('operator', 'or')
        if operator not in ['and', 'or']:
            raise ValueError(_('malformed requirements for option: "{0}"'
                               ' operator must be "or" or "and"').format(operator))
        return operator


    ret_requires = {}
    config_action = set()

    # start parsing all requires given by user (has dict)
    # transforme it to a tuple
    for require in requires:
        if not isinstance(require, dict):
            raise ValueError(_("malformed requirements type for option:"
                               " {0}, must be a dict").format(name))
        valid_keys = ('option', 'expected', 'action', 'inverse', 'transitive',
                      'same_action', 'operator')
        unknown_keys = frozenset(require.keys()) - frozenset(valid_keys)
        if unknown_keys != frozenset():
            raise ValueError(_('malformed requirements for option: {0}'
                             ' unknown keys {1}, must only '
                             '{2}').format(name,
                                           unknown_keys,
                                           valid_keys))
        # prepare all attributes
        if not ('expected' in require and isinstance(require['expected'], list)) and \
                not ('option' in require and 'expected' in require) or \
                'action' not in require:
            raise ValueError(_("malformed requirements for option: {0}"
                               " require must have option, expected and"
                               " action keys").format(name))
        action = get_action(require)
        config_action.add(action)
        if action not in ret_requires:
            ret_requires[action] = {}
        set_expected(require, ret_requires)

    # transform dict to tuple
    ret = []
    for requires in ret_requires.values():
        ret_action = []
        for require in requires.values():
            ret_action.append((tuple(require[0]), require[1],
                               require[2], require[3], require[4], require[5]))
        ret.append(tuple(ret_action))
    return frozenset(config_action), tuple(ret)


class SymLinkOption(OnlyOption):

    def __init__(self, name, opt):
        if not isinstance(opt, Option):
            raise ValueError(_('malformed symlinkoption '
                               'must be an option '
                               'for symlink {0}').format(name))
        _setattr = object.__setattr__
        _setattr(self, '_name', name)
        _setattr(self, '_opt', opt)
        opt._set_has_dependency()

    def __getattr__(self, name, context=undefined):
        return getattr(self._impl_getopt(), name)

    def _impl_getopt(self):
        return self._opt

    def impl_get_information(self, key, default=undefined):
        return self._impl_getopt().impl_get_information(key, default)

    def impl_is_readonly(self):
        return True

    def impl_getproperties(self):
        return self._impl_getopt()._properties

    def impl_get_callback(self):
        return self._impl_getopt().impl_get_callback()

    def impl_has_callback(self):
        "to know if a callback has been defined or not"
        return self._impl_getopt().impl_has_callback()

    def impl_is_multi(self):
        return self._impl_getopt().impl_is_multi()

    def _is_subdyn(self):
        return getattr(self._impl_getopt(), '_subdyn', None) is not None

    def _get_consistencies(self):
        return ()

    def _has_consistencies(self):
        return False


class DynSymLinkOption(object):
    __slots__ = ('_dyn', '_opt', '_name')

    def __init__(self, name, opt, dyn):
        self._name = name
        self._dyn = dyn
        self._opt = opt

    def __getattr__(self, name, context=undefined):
        return getattr(self._impl_getopt(), name)

    def impl_getname(self):
        return self._name

    def impl_get_display_name(self):
        return self._impl_getopt().impl_get_display_name(dyn_name=self.impl_getname())

    def _impl_getopt(self):
        return self._opt

    def impl_getsuffix(self):
        return self._dyn.split('.')[-1][len(self._impl_getopt().impl_getname()):]

    def impl_getpath(self, context):
        path = self._impl_getopt().impl_getpath(context)
        base_path = '.'.join(path.split('.')[:-2])
        if self.impl_is_master_slaves() and base_path is not '':
            base_path = base_path + self.impl_getsuffix()
        if base_path == '':
            return self._dyn
        else:
            return base_path + '.' + self._dyn

    def impl_validate(self, value, context=undefined, validate=True,
                      force_index=None, force_submulti_index=None, is_multi=None,
                      display_error=True, display_warnings=True, multi=None,
                      setting_properties=undefined):
        return self._impl_getopt().impl_validate(value, context, validate,
                                                 force_index,
                                                 force_submulti_index,
                                                 current_opt=self,
                                                 is_multi=is_multi,
                                                 display_error=display_error,
                                                 display_warnings=display_warnings,
                                                 multi=multi,
                                                 setting_properties=setting_properties)

    def impl_is_dynsymlinkoption(self):
        return True
